﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;

namespace Reversi
{
    // 着手データクラス
    class Move
    {
        // 着手のパス
        public static Move PASS = new Move(-1, -1);

        // 着手の位置
        public int x;
        public int y;

        // コンストラクタ(位置を指定)
        public Move(int x, int y)
        {
            this.x = x;
            this.y = y;
        }

        public bool equal(Move move)
        {
            return ((this.x == move.x) && (this.y == move.y)) ? true : false;
        }
    }

    // ひっくり返しフラグクラス
    class ReverseFlag
    {
        // 着手した石
        public const int MOVE_STONE = 1;
        // ひっくり返った石
        public const int REVERSED_STONE = 2;

        // 盤面
        public int[,] stone = new int[Board.SIZE, Board.SIZE];
        
        // リセット
        public void reset()
        {
            for (int x = 0; x < Board.SIZE; x++)
            {
                for (int y = 0; y < Board.SIZE; y++)
                {
                    stone[x, y] = 0;
                }
            }
        }
    }

    // 盤面クラス
    class Board
    {
        // 先手
        public const int FIRST_MOVE = 1;
        // 後手
        public const int SECOND_MOVE = -1;
        // 石が無い
        public const int NO_STONE = 0;
        // 引き分け
        public const int DRAW = 0;
        // 路の数 (通常の8路盤)
        public const int SIZE = 8;

        // 盤面
        public int[,] stone = new int[SIZE, SIZE];
        // 対局開始からの手順数
        public int turn;

        // 手番持ちのプレイヤー（先手 or 後手）
        public int turnHolder
        {
            get
            {
                return ((turn % 2) == 0) ? FIRST_MOVE : SECOND_MOVE;
            }
        }

        // 勝者
        public int winner
        {
            get
            {
                int f = countStone(FIRST_MOVE);
                int s = countStone(SECOND_MOVE);

                if (f > s)
                {
                    return FIRST_MOVE;
                }
                else if(f < s)
                {
                    return SECOND_MOVE;
                }
                else
                {
                    return DRAW;
                }
            }
        }

        // 盤面のコピーを生成する
        public Board copy()
        {
            Board board = new Board();

            for (int x = 0; x < SIZE; x++)
            {
                for (int y = 0; y < SIZE; y++)
                {
                    board.stone[x, y] = this.stone[x, y];
                }
            }
            board.turn = this.turn;

            return board;
        }

        // 盤面の範囲内か判定する
        private static bool isInBoard(int x, int y)
        {
            if ((x < 0) || (x >= SIZE)) return false;
            if ((y < 0) || (y >= SIZE)) return false;
            return true;
        }

        // 盤面を初期状態にリセットする
        public void reset()
        {
            for (int x = 0; x < SIZE; x++)
            {
                for (int y = 0; y < SIZE; y++)
                {
                    stone[x, y] = NO_STONE;
                }
            }
            stone[SIZE / 2 - 1, SIZE / 2 - 1] = SECOND_MOVE;
            stone[SIZE / 2 - 1, SIZE / 2    ] = FIRST_MOVE;
            stone[SIZE / 2    , SIZE / 2 - 1] = FIRST_MOVE;
            stone[SIZE / 2    , SIZE / 2    ] = SECOND_MOVE;

            turn = 0;
        }
        
        // 着手可能な手があるか？
        public bool hasAvailableMove()
        {
            List<Reversi.Move> nextMoves = enumNextMoves();
            if (nextMoves.Count == 1)
            {
                if (nextMoves[0].equal(Reversi.Move.PASS))
                {
                    return false;
                }
            }
            return true;
        }

        // 有効な着手か判定する
        public bool isAvailableMove(Move move)
        {
            // 彼我のどちらが先手/後手か
            int mine = turnHolder;
            int his = (mine == FIRST_MOVE) ? SECOND_MOVE : FIRST_MOVE;

            // 範囲内で石の無い位置であることをチェック
            if (!isInBoard(move.x, move.y)) return false;
            if (stone[move.x, move.y] != NO_STONE) return false;

            // 八方の石をチェック
            for (int dx = -1; dx <= 1; dx++)
            {
                for (int dy = -1; dy <= 1; dy++)
                {
                    if ((dx == 0) && (dy == 0)) continue;

                    for (int i = 1; i < SIZE; i++)
                    {
                        int x = move.x + i * dx;
                        int y = move.y + i * dy;
                        if (!isInBoard(x, y)) break;
                        int s = stone[x, y];
                        if (s == NO_STONE) break;
                        else if (s == his) continue;
                        else if (s == mine)
                        {
                            if (i == 1) break;
                            return true;
                        }
                        else break; // abnormal case
                    }
                }
            }
            return false;
        }

        // 着手する
        public ReverseFlag doMove(Move move)
        {
            ReverseFlag reverse = new ReverseFlag();

            // 彼我のどちらが先手/後手か
            int mine = turnHolder;
            int his = (mine == FIRST_MOVE) ? SECOND_MOVE : FIRST_MOVE;
            // 手順数を進める
            turn++;

            // 範囲内で石の無い位置であることをチェック
            if (!isInBoard(move.x, move.y)) return reverse;
            if (stone[move.x, move.y] != NO_STONE) return reverse;

            // 着手
            stone[move.x, move.y] = mine;
            reverse.stone[move.x, move.y] = ReverseFlag.MOVE_STONE;

            // 八方の石をチェック
            for (int dx = -1; dx <= 1; dx++)
            {
                for (int dy = -1; dy <= 1; dy++)
                {
                    if ((dx == 0) && (dy == 0)) continue;

                    for (int i = 1; i < SIZE; i++)
                    {
                        int x = move.x + i * dx;
                        int y = move.y + i * dy;
                        if (!isInBoard(x, y)) break;
                        int s = stone[x, y];
                        if (s == NO_STONE) break;
                        else if (s == his) continue;
                        else if (s == mine)
                        {
                            // ひっくり返す
                            for (int j = 1; j < i; j++)
                            {
                                stone[move.x + j * dx, move.y + j * dy] = mine;
                                reverse.stone[move.x + j * dx, move.y + j * dy] = ReverseFlag.REVERSED_STONE;
                            }
                            break;
                        }
                        else break; // abnormal case
                    }
                }
            }
            return reverse;
        }

        // 石の数を数える
        public int countStone(int player)
        {
            int count = 0;

            for (int x = 0; x < SIZE; x++)
            {
                for (int y = 0; y < SIZE; y++)
                {
                    if(stone[x,y] == player) count++;
                }
            }
            return count;
        }

        // 対局終了判定
        public bool hasFinished()
        {
            // 盤が石で埋まれば終了
            if (countStone(NO_STONE) == 0)
            {
                return true;
            }
            // 片方の石がゼロになれば終了
            if(countStone(FIRST_MOVE) == 0 || countStone(SECOND_MOVE) == 0)
            {
                return true;
            }
            // 双方着手不可能な場合
            if (!hasAvailableMove())
            {
                Board nextBoard = this.copy();
                nextBoard.doMove(Move.PASS);
                if(!nextBoard.hasAvailableMove())
                {
                    return true;
                }
            }
            return false;
        }

        // 次の局面を列挙する
        public List<Move> enumNextMoves()
        {
            List<Move> nextMoves = new List<Move>();

            // 盤面から有効な着手を探す
            for (int x = 0; x < SIZE; x++)
            {
                for (int y = 0; y < SIZE; y++)
                {
                    Move move = new Move(x, y);
                    if (isAvailableMove(move))
                    {
                        // 有効な着手と、その場合の局面を記憶
                        nextMoves.Add(move);
                    }
                }
            }
            // 有効な着手が1つも無い場合は、パスの一手のみを記憶
            if (nextMoves.Count == 0)
            {
                nextMoves.Add(Move.PASS);
            }
            return nextMoves;
        }
    }
}
